import pytest
from pytest_lazyfixture import lazy_fixture
import stk
import stko

from .case_data import CaseData


@pytest.fixture
def case_data_1(atomic_number, id, charge, position):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        atom=stko.PositionedAtom(
            atom=stk.Atom(
                id=id,
                atomic_number=atomic_number,
                charge=charge,
            ),
            position=position,
        ),
        id=id,
        charge=charge,
        atomic_number=atomic_number,
        position=position,
    )


@pytest.fixture(
    params=(
        lazy_fixture('case_data_1'),
    )
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param


@pytest.fixture(
    params=[0, 3],
)
def id(request):
    """
    An atom id.

    """

    return request.param


@pytest.fixture(
    params=[0],
)
def charge(request):
    """
    An atomic charge.

    """

    return request.param


@pytest.fixture(params=tuple(range(1, 118)))
def atomic_number(request):
    """
    An atomic number.

    """

    return request.param


@pytest.fixture(
    params=[(0, 0, 0), (100, 0, -1)],
)
def position(request):
    """
    A position.

    """

    return request.param


@pytest.fixture(
    params=[
        cls for cls in stk.__dict__.values()
        if isinstance(cls, type)
        and issubclass(cls, stk.AtomImpl)
        and cls is not stk.AtomImpl
    ],
)
def cls(request):
    """
    Return an :class:`.Atom` instance.

    Parameters
    ----------
    id : :class:`int`
        The id of the returned atom.

    charge : :class:`float`
        The charge of the returned atom.

    Returns
    -------
    :class:`.Atom`
        An atom.

    """

    return request.param


@pytest.fixture
def positioned_atom(cls, position):
    """
    An :class:`.PositionedAtom` instance.

    """

    return stko.PositionedAtom(
        atom=cls(3, -5),
        position=position,
    ).clone()



"""
Molecule Splitter
=================

#. :class:`.MoleculeSplitter`

Class for splitting a molecule into many with new connectors.

"""

import logging
from rdkit.Chem import AllChem as rdkit
from itertools import combinations

import stk

from ..atoms import Du

logger = logging.getLogger(__name__)


class MoleculeSplitter:
    """
    Split an stk.molecule into many with dummy atoms.

    Examples
    --------

    Given a molecule, this class allows you to break bonds based on
    `breaker_smarts` between the atoms in `bond_deleter_ids`.

    .. code-block:: python

        import stk
        import stko

        full_mol = stk.BuildingBlock('C1=CC=NC(=C1)C=NC2=CC=C(C=C2)Br')

        splitter = stko.MoleculeSplitter(
            breaker_smarts='[#6X3]~[#7X2]~[#6X3H1]~[#6X3!H1]',
            bond_deleter_ids=(0, 1),
        )
        split_mols = splitter.split(full_mol)

    """

    def __init__(
        self,
        breaker_smarts,
        bond_deleter_ids,
    ):
        """
        Initialize a :class:`.MoleculeSplitter`.

        Parameters
        ----------
        breaker_smarts : :class:`str`
            SMARTS string used to find the substructure to break.

        bond_deleter_ids : :class:`tuple` of :class:`int`
            Index of atoms in `breaker_smarts` to break bond between.

        """

        self._breaker_smarts = breaker_smarts
        self._bond_deleter_ids = bond_deleter_ids

    def split(self, molecule):
        """
        Split a molecule.

        Parameters
        ----------
        molecule : :class:`stk.Molecule`
            Molecule to modify.

        Returns
        -------
        molecules : :class:`iterable` of :class:`stk.BuildingBlock`
            The resulting list of molecules.

        """

        rdkit_mol = molecule.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_mol)
        bond_pair_ids_to_delete = []
        reactable_atom_ids = []
        for atom_ids in rdkit_mol.GetSubstructMatches(
            query=rdkit.MolFromSmarts(self._breaker_smarts)
        ):
            # Get the atom ids of the query on either end of the broken
            # bond, translated from query ids to molecule atom ids.
            translated_bond_atom_ids = (
                atom_ids[self._bond_deleter_ids[0]],
                atom_ids[self._bond_deleter_ids[1]],
            )
            # Get tuples of the bonds to break by atom ids.
            bond_pair_ids_to_delete.extend(
                tuple(sorted((i, j)))
                for i, j in combinations(translated_bond_atom_ids, 2)
            )
            # Get the atom ids of the atoms on either end of the broken
            # bond at which reactions can occur.
            reactable_atom_ids.extend(
                i for i in translated_bond_atom_ids
            )

        # Get the rdkit molecule bond ids associated with the bonds to
        # delete.
        bond_ids_to_delete = []
        for bond in rdkit_mol.GetBonds():
            idxs = tuple(sorted((
                bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            )))
            if idxs in bond_pair_ids_to_delete:
                bond_ids_to_delete.append(bond.GetIdx())

        # Delete bonds and atoms.
        fragments = rdkit.GetMolFrags(
            rdkit.FragmentOnBonds(
                mol=rdkit_mol,
                bondIndices=tuple(bond_ids_to_delete),
                addDummies=True,
            ),
            asMols=True,
        )

        molecules = []
        for frag in fragments:
            atoms = []
            for a in frag.GetAtoms():
                if a.GetAtomicNum() == 0:
                    atoms.append(Du(a.GetIdx()))
                else:
                    atoms.append(
                        stk.Atom(
                            id=a.GetIdx(),
                            atomic_number=a.GetAtomicNum(),
                            charge=a.GetFormalCharge(),
                        )
                    )
            atoms = tuple(atoms)

            bonds = tuple(
                stk.Bond(
                    atom1=atoms[b.GetBeginAtomIdx()],
                    atom2=atoms[b.GetEndAtomIdx()],
                    order=(
                        9 if b.GetBondType() == rdkit.BondType.DATIVE
                        else b.GetBondTypeAsDouble()
                    )
                )
                for b in frag.GetBonds()
            )
            position_matrix = frag.GetConformer().GetPositions()
            molecules.append(
                stk.BuildingBlock.init(
                    atoms=atoms,
                    bonds=bonds,
                    position_matrix=position_matrix,
                )
            )

        return molecules



"""
Orca Extractor
=============

#. :class:`.OrcaExtractor`

Class to extract properties from Orca output.

"""

import re
from .extractor import Extractor


class OrcaExtractor(Extractor):
    """
    Extracts properties from Orca 4.2 output files.

    Limited to final single point energy for now.

    Attributes
    ----------
    output_file : :class:`str`
        Output file to extract properties from.

    output_lines : :class:`list` : :class:`str`
        :class:`list` of all lines in as :class:`str` in the output
        file.

    total_energy : :class:`float`
        The total energy in the :attr:`output_file` as
        :class:`float`. The energy is in units of a.u..

    """

    def _extract_values(self):
        """
        Extract all properties from Orca output file.

        Returns
        -------
        None : :class:`NoneType`

        """

        for i, line in enumerate(self.output_lines):
            if self._check_line(line, 'total_energy'):
                self._extract_total_energy(line)

    def _properties_dict(self):

        return {'total_energy': 'FINAL SINGLE POINT ENERGY'}

    def _extract_total_energy(self, line):
        """
        Updates :attr:`total_energy`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.total_energy = float(string)



"""
XTB Calculators
===============

#. :class:`.XTBEnergy`

Wrappers for calculators within the :mod:`xtb` code.

"""

import logging
import os
import shutil
import uuid
import subprocess as sp

from .calculators import Calculator
from .results import XTBResults
from ..utilities import is_valid_xtb_solvent, XTBInvalidSolventError

logger = logging.getLogger(__name__)


class XTBEnergy(Calculator):
    """
    Uses GFN-xTB [1]_ to calculate energy and other properties.

    By default, :meth:`get_results` will extract other properties of
    the :class:`.Molecule` passed to :meth:`calculate`, which
    will be saved in the attributes of :class:`.XTBResults`.

    Notes
    -----
    When running :meth:`calculate`, this calculator changes the
    present working directory with :func:`os.chdir`. The original
    working directory will be restored even if an error is raised, so
    unless multi-threading is being used this implementation detail
    should not matter.

    If multi-threading is being used an error could occur if two
    different threads need to know about the current working directory
    as :class:`.XTBEnergy` can change it from under them.

    Note that this does not have any impact on multi-processing,
    which should always be safe.

    *Contributors*
    We thank Andrew Tarzia and Alejandro Santana-Bonilla for their
    contributions to this code.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('NCCNCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit="AB",
                orientations=[0, 0],
                num_repeating_units=1
            )
        )

        # Optimize the constructed molecule so that it has a
        # reasonable structure.
        opt = stko.OptimizerSequence(
            stko.UFF(),
            stko.XTB(xtb_path='/opt/gfnxtb/xtb', unlimited_memory=True)
        )
        polymer = opt.optimize(polymer)

        # Calculate energy using GFN-xTB.
        xtb = stko.XTBEnergy(
            xtb_path='/opt/gfnxtb/xtb',
            unlimited_memory=True
        )

        xtb_results = xtb.get_results(polymer)

        # Extract properties from the energy calculator for a given
        # molecule.
        total_energy = xtb_results.get_total_energy()
        homo_lumo_gap = xtb_results.get_homo_lumo_gap()
        fermi_levels = xtb_results.get_fermi_level()
        homo_lumo_orbitals = xtb_results.get_homo_lumo_orbitals()
        full_dipole_moments = xtb_results.get_full_dipole_moments()

    If `calculate_free_energy` is ``True``, xTB performs a
    numerical Hessian calculation and calculates the total free energy
    and vibrational frequencies of a molecule. It is recommended that a
    well optimized structure be used as input for these calculations

    .. code-block:: python

        # Optimize the constructed molecule so that it has a
        # reasonable structure.
        optimizer = stko.OptimizerSequence(
            stko.ETKDG(),
            stko.XTB(
                xtb_path='/opt/gfnxtb/xtb',
                unlimited_memory=True,
                opt_level='verytight'
            )
        )
        polymer = optimizer.optimize(polymer)

        # Calculate energy using GFN-xTB.
        xtb = stko.XTBEnergy(
            xtb_path='/opt/gfnxtb/xtb',
            unlimited_memory=True,
            calculate_free_energy=True
        )

        xtb_results = xtb.get_results(polymer)

        # Extract properties from the energy calculator for a given
        # molecule.
        total_free_energy = xtb_results.get_total_free_energy()
        total_frequencies = xtb_results.get_frequencies()

    If `calculate_ip_and_ea` is ``True``, xTB performs multiple single-
    point-energy calculations to calculate the vertical electron
    affinity and ionisation potential. It is recommended that a
    well optimized structure be used as input for these calculations

    .. code-block:: python

        # Optimize the constructed molecule so that it has a
        # reasonable structure.
        optimizer = stko.OptimizerSequence(
            stko.ETKDG(),
            stko.XTB(
                xtb_path='/opt/gfnxtb/xtb',
                unlimited_memory=True,
                opt_level='verytight'
            )
        )
        polymer = optimizer.optimize(polymer)

        # Calculate energy using GFN-xTB.
        xtb = stko.XTBEnergy(
            xtb_path='/opt/gfnxtb/xtb',
            unlimited_memory=True,
            calculate_ip_and_ea=True,
        )

        xtb_results = xtb.get_results(polymer)

        # Extract properties from the energy calculator for a given
        # molecule.
        ip = xtb_results.get_ionisation_potential()
        ea = xtb_results.get_electron_affinity()

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/setup.html

    """
    def __init__(
        self,
        xtb_path,
        gfn_version=2,
        output_dir=None,
        num_cores=1,
        calculate_free_energy=False,
        calculate_ip_and_ea=False,
        electronic_temperature=300,
        solvent_model='gbsa',
        solvent=None,
        solvent_grid='normal',
        charge=0,
        num_unpaired_electrons=0,
        unlimited_memory=False,
    ):
        """
        Initializes a :class:`XTBEnergy` instance.

        Parameters
        ----------
        xtb_path : :class:`str`
            The path to the xTB executable.

        gfn_version : :class:`int`, optional
            Parameterization of GFN to use in xTB.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/basics.html.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the calculation are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        num_cores : :class:`int`, optional
            The number of cores xTB should use.

        calculate_free_energy : :class:`bool`, optional
            Whether to calculate the total free energy and vibrational
            frequencies. Setting this to ``True`` can drastically
            increase calculation time and memory requirements.

        calculate_ip_and_ea : :class:`bool`, optional
            Whether to calculate the vertical ionisation potential and
            vertical electron affinity. Equivalent to  `--vipea` flag
            in command-line interface.

        electronic_temperature : :class:`int`, optional
            Electronic temperature in Kelvin.

        solvent_model : :class:`str`
            Solvent model to use out of older `gbsa` and newer `alpb`.
            `gbsa` is default for backwards compatability, but `alpb`
            is recommended.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        solvent : :class:`str`, optional
            Solvent to use in GBSA implicit solvation method.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        solvent_grid : :class:`str`, optional
            Grid level to use in SASA calculations for GBSA implicit
            solvent.
            Can be one of ``'normal'``, ``'tight'``, ``'verytight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        charge : :class:`int`, optional
            Formal molecular charge.

        num_unpaired_electrons : :class:`int`, optional
            Number of unpaired electrons.

        unlimited_memory : :class: `bool`, optional
            If ``True`` :meth:`energy` will be run without constraints
            on the stack size. If memory issues are encountered, this
            should be ``True``, however this may raise issues on
            clusters.

        """

        if solvent is not None:
            solvent = solvent.lower()
            if gfn_version == 0:
                raise XTBInvalidSolventError(
                    'No solvent valid for version',
                    f' {gfn_version!r}.'
                )
            if not is_valid_xtb_solvent(
                gfn_version=gfn_version,
                solvent_model=solvent_model,
                solvent=solvent,
            ):
                raise XTBInvalidSolventError(
                    f'Solvent {solvent!r} and model {solvent_model!r}',
                    f' is invalid for version {gfn_version!r}.'
                )

        self._xtb_path = xtb_path
        self._gfn_version = str(gfn_version)
        self._output_dir = output_dir
        self._num_cores = str(num_cores)
        self._calculate_free_energy = calculate_free_energy
        self._calculate_ip_and_ea = calculate_ip_and_ea
        self._electronic_temperature = str(electronic_temperature)
        self._solvent = solvent
        self._solvent_model = solvent_model
        self._solvent_grid = solvent_grid
        self._charge = str(charge)
        self._num_unpaired_electrons = str(num_unpaired_electrons)
        self._unlimited_memory = unlimited_memory

    def _write_detailed_control(self):
        string = f'$gbsa\n   gbsagrid={self._solvent_grid}'

        with open('det_control.in', 'w') as f:
            f.write(string)

    def _run_xtb(self, xyz, out_file, init_dir, output_dir):
        """
        Runs GFN-xTB.

        Parameters
        ----------
        xyz : :class:`str`
            The name of the input structure ``.xyz`` file.

        out_file : :class:`str`
            The name of output file with xTB results.

        init_dir : :class:`str`
            The name of the current working directory.

        output_dir : :class:`str`
            The name of the directory into which files generated during
            the calculation are written.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Modify the memory limit.
        if self._unlimited_memory:
            memory = 'ulimit -s unlimited ;'
        else:
            memory = ''

        if self._solvent is not None:
            solvent = f'--{self._solvent_model} {self._solvent} '
        else:
            solvent = ''

        if self._calculate_free_energy:
            hess = '--hess'
        else:
            hess = ''

        if self._calculate_ip_and_ea:
            vipea = '--vipea'
        else:
            vipea = ''

        cmd = (
            f'{memory} {self._xtb_path} '
            f'{xyz} --gfn {self._gfn_version} '
            f'{hess} {vipea} --parallel {self._num_cores} '
            f'--etemp {self._electronic_temperature} '
            f'{solvent} --chrg {self._charge} '
            f'--uhf {self._num_unpaired_electrons} -I det_control.in'
        )

        try:
            os.chdir(output_dir)
            self._write_detailed_control()
            with open(out_file, 'w') as f:
                # Note that sp.call will hold the program until
                # completion of the calculation.
                sp.call(
                    cmd,
                    stdin=sp.PIPE,
                    stdout=f,
                    stderr=sp.PIPE,
                    # Shell is required to run complex arguments.
                    shell=True
                )
        finally:
            os.chdir(init_dir)

    def calculate(self, mol):
        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        os.mkdir(output_dir)

        init_dir = os.getcwd()
        xyz = os.path.join(output_dir, 'input_structure.xyz')
        out_file = os.path.join(output_dir, 'energy.output')
        mol.write(xyz)
        yield self._run_xtb(
            xyz=xyz,
            out_file=out_file,
            init_dir=init_dir,
            output_dir=output_dir,
        )

    def get_results(self, mol):
        """
        Calculate the xTB properties of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`.XTBResults`
            The properties, with units, from xTB calculations.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        out_file = os.path.join(output_dir, 'energy.output')

        return XTBResults(
            generator=self.calculate(mol),
            output_file=out_file,
        )

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        """

        return self.get_results(mol).get_total_energy()[0]



import stk
import stko

from ..utilities import is_equivalent_atom


def test_repr(positioned_atom):
    """
    Test :meth:`.PositionedAtom.__repr__`.

    Parameters
    ----------
    atom : :class:`.PositionedAtom`
        The atom, whose representation should be tested.

    Returns
    -------
    None : :class:`NoneType`

    """

    other = stko.PositionedAtom(
        atom=eval(repr(positioned_atom), dict(stk.__dict__)),
        position=positioned_atom.get_position(),
    )
    is_equivalent_atom(other, positioned_atom)



import pytest
import numpy as np
import stk


class CaseData:
    """
    A test case.

    Attributes:
        molecule:
            The molecule to be tested.

        unoptimised_energy:
            The energy of the molecule from stk generation.

    """

    position_matrix: np.ndarray

    def __init__(self, molecule, unoptimised_energy):

        self.molecule = molecule
        self.unoptimised_energy = unoptimised_energy


@pytest.fixture(
    scope='session',
    params=[
        CaseData(
            molecule=stk.BuildingBlock('NCCN'),
            unoptimised_energy=98.68403256936926,
        ),
        CaseData(
            molecule=stk.BuildingBlock(
                'C(#Cc1cccc2ccncc21)c1ccc2[nH]c3ccc(C#Cc4cccc5cnccc54)'
                'cc3c2c1'
            ),
            unoptimised_energy=2032.8111567743895,
        ),
        CaseData(
            molecule=stk.BuildingBlock('CCCCCC'),
            unoptimised_energy=141.44622279628743,
        ),
        CaseData(
            molecule=stk.BuildingBlock('c1ccccc1'),
            unoptimised_energy=56.73128282588534,
        ),
        CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCNCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=2,
                    optimizer=stk.MCHammer(),
                ),
            ),
            unoptimised_energy=15002.946293854524,
        ),
    ],
)
def case_molecule(request):
    """
    A :class:`.Molecule` instance.

    """

    return request.param



"""
OpenBabel Calculators
=====================

#. :class:`.OpenBabelEnergy`

Wrappers for calculators within the `openbabel` code.

"""

import logging
import os
from openbabel import openbabel

from .calculators import Calculator
from .results import EnergyResults


logger = logging.getLogger(__name__)


class OpenBabelEnergy(Calculator):
    """
    Uses OpenBabel to calculate forcefield energies.[1]_

    Examples
    --------
    .. code-block:: python

        import stk
        import stko

        # Create a molecule whose energy we want to know.
        mol1 = stk.BuildingBlock('CCCNCCCN')

        # Create the energy calculator.
        energy_calc = stko.OpenBabelEnergy('uff')

        # Calculate the energy.
        results = energy_calc.get_results(mol1)
        energy = results.get_energy()
        unit_string = results.get_unit_string()

    References
    ----------
    .. [1] http://openbabel.org/dev-api/classOpenBabel_1_
    1OBForceField.shtml#a2f2732698efde5c2f155bfac08fd9ded

    """

    def __init__(self, forcefield):
        """
        Initialize `openbabel` forcefield energy calculation.

        Parameters
        ----------
        forcefield : :class:`str`
            Forcefield to use. Options include `uff`, `gaff`,
            `ghemical`, `mmff94`.

        """

        self._forcefield = forcefield

    def calculate(self, mol):
        temp_file = 'temp.mol'
        mol.write(temp_file)
        obConversion = openbabel.OBConversion()
        obConversion.SetInFormat("mol")
        OBMol = openbabel.OBMol()
        obConversion.ReadFile(OBMol, temp_file)
        OBMol.PerceiveBondOrders()
        os.system('rm temp.mol')

        forcefield = openbabel.OBForceField.FindForceField(
            self._forcefield
        )
        forcefield.Setup(OBMol)

        yield forcefield.Energy()

    def get_results(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`.EnergyResults`
            The energy and units of the energy.

        """

        return EnergyResults(
            generator=self.calculate(mol),
            unit_string='kJ mol-1',
        )

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        """

        return self.get_results(mol).get_energy()



import setuptools

setuptools.setup(
    name="stko",
    version="0.0.38",
    author="Steven Bennett, Andrew Tarzia",
    author_email="s.bennett18@imperial.ac.uk",
    description=(
        "Contains molecular optimizers and property calculators for "
        "use with stk."
    ),
    url="https://github.com/JelfsMaterialsGroup/stko",
    packages=setuptools.find_packages(),
    install_requires=(
        'scipy',
        'numpy',
        'networkx',
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
Shape Results
=============

#. :class:`.ShapeResults`

Results class for extracting molecular shape properties.

"""

from .results import Results


class ShapeResults(Results):
    """
    Results class containing molecule shape measures.

    """

    def __init__(self, generator):

        self._values = next(generator)

    def get_pmi1(self):
        return self._values['pmi1']

    def get_pmi2(self):
        return self._values['pmi2']

    def get_pmi3(self):
        return self._values['pmi3']

    def get_npr1(self):
        return self._values['npr1']

    def get_npr2(self):
        return self._values['npr2']

    def get_asphericity(self):
        return self._values['asphericity']

    def get_eccentricity(self):
        return self._values['eccentricity']

    def get_inertial_shape_factor(self):
        return self._values['inertialshapefactor']

    def get_radius_of_gyration(self):
        return self._values['radiusofgyration']

    def get_spherocity_index(self):
        return self._values['spherocityindex']



"""
MD Analysis
===========

#. :class:`.MDAnalysis`

Class for converting a molecule to and back from an MDAnalysis object.

"""

import logging

import MDAnalysis as mda

logger = logging.getLogger(__name__)


class MDAnalysis:
    """
    Converter for :class:`stk.Molecule` to and from MDAnalysis.

    Examples
    --------

    An stk molecule can be converted into an MDAnalysis Universe.

    .. code-block:: python

        import stk
        import stko

        stkmol = stk.BuildingBlock('NCCNCCN').with_centroid(
            position=np.array((10, 10, 10))
        )
        universe = stko.MDAnalysis().get_universe(stkmol)

        print('R_g:', universe.atoms.radius_of_gyration())
        print('B_sphere:', universe.atoms.bsphere())
        print('Universe COM:', universe.atoms.center_of_mass())
        print('stk centroid:', stkmol.get_centroid())

    """

    def get_universe(self, mol):
        """
        Get an MDAnalysis object.

        Parameters
        ----------
        mol : :class:`stk.Molecule`
            Molecule to convert.

        Returns
        -------
        :class:`MDAnalysis.Universe`
            The MDAnalysis Universe of the molecule.

        """

        rdkit_mol = mol.to_rdkit_mol()
        return mda.Universe(rdkit_mol)



import stko
from ..utilities import (
    inequivalent_position_matrices,
    is_equivalent_molecule,
)
import numpy as np


def test_MMFF_opt(case_mmff_molecule):
    energy = (
        stko.MMFFEnergy().get_energy(case_mmff_molecule.molecule)
    )
    optimizer = stko.MMFF()
    opt_res = optimizer.optimize(case_mmff_molecule.molecule)
    is_equivalent_molecule(opt_res, case_mmff_molecule.molecule)
    inequivalent_position_matrices(
        opt_res, case_mmff_molecule.molecule
    )
    opt_energy = (
        stko.MMFFEnergy().get_energy(opt_res)
    )
    assert np.isclose(
        energy, case_mmff_molecule.unoptimised_energy, atol=1E-3
    )
    assert opt_energy < case_mmff_molecule.unoptimised_energy


def test_UFF_opt(case_uff_molecule):
    energy = (
        stko.UFFEnergy().get_energy(case_uff_molecule.molecule)
    )
    optimizer = stko.UFF()
    opt_res = optimizer.optimize(case_uff_molecule.molecule)
    is_equivalent_molecule(opt_res, case_uff_molecule.molecule)
    inequivalent_position_matrices(opt_res, case_uff_molecule.molecule)
    opt_energy = (
        stko.UFFEnergy().get_energy(opt_res)
    )
    assert np.isclose(
        energy, case_uff_molecule.unoptimised_energy, atol=1E-3
    )
    assert opt_energy < case_uff_molecule.unoptimised_energy


def test_ETKDG_opt(case_etkdg_molecule):
    optimizer = stko.ETKDG()
    opt_res = optimizer.optimize(case_etkdg_molecule)
    is_equivalent_molecule(opt_res, case_etkdg_molecule)
    inequivalent_position_matrices(opt_res, case_etkdg_molecule)

    optimizer = stko.ETKDG(random_seed=2584)
    opt_res = optimizer.optimize(case_etkdg_molecule)
    is_equivalent_molecule(opt_res, case_etkdg_molecule)
    inequivalent_position_matrices(opt_res, case_etkdg_molecule)



"""
This module defines general-purpose objects, functions and classes.

Functions, classes, etc. defined here should not depend on any other
part of ``stko``. They must be completely self-sufficient.

"""

import rdkit.Chem.AllChem as rdkit
from rdkit.Geometry import Point3D
import numpy as np
import shutil
import os
import subprocess as sp
import gzip
import re
from collections import deque, defaultdict
from glob import iglob
from itertools import chain
from scipy.spatial.distance import euclidean


# This dictionary gives easy access to the rdkit bond types.
bond_dict = {'1': rdkit.rdchem.BondType.SINGLE,
             'am': rdkit.rdchem.BondType.SINGLE,
             '2': rdkit.rdchem.BondType.DOUBLE,
             '3': rdkit.rdchem.BondType.TRIPLE,
             'ar': rdkit.rdchem.BondType.AROMATIC}

# A dictionary which matches atomic number to elemental symbols.
periodic_table = {
    1: 'H', 2: 'He', 3: 'Li', 4: 'Be', 5: 'B', 6: 'C',
    7: 'N', 8: 'O', 9: 'F', 10: 'Ne', 11: 'Na', 12: 'Mg',
    13: 'Al', 14: 'Si', 15: 'P', 16: 'S', 17: 'Cl',
    18: 'Ar', 19: 'K', 20: 'Ca', 21: 'Sc', 22: 'Ti',
    23: 'V', 24: 'Cr', 25: 'Mn', 26: 'Fe', 27: 'Co',
    28: 'Ni', 29: 'Cu', 30: 'Zn', 31: 'Ga', 32: 'Ge',
    33: 'As', 34: 'Se', 35: 'Br', 36: 'Kr', 37: 'Rb',
    38: 'Sr', 39: 'Y', 40: 'Zr', 41: 'Nb', 42: 'Mo',
    43: 'Tc', 44: 'Ru', 45: 'Rh', 46: 'Pd', 47: 'Ag',
    48: 'Cd', 49: 'In', 50: 'Sn', 51: 'Sb', 52: 'Te',
    53: 'I', 54: 'Xe', 55: 'Cs', 56: 'Ba', 57: 'La',
    58: 'Ce', 59: 'Pr', 60: 'Nd', 61: 'Pm', 62: 'Sm',
    63: 'Eu', 64: 'Gd', 65: 'Tb', 66: 'Dy', 67: 'Ho',
    68: 'Er', 69: 'Tm', 70: 'Yb', 71: 'Lu', 72: 'Hf',
    73: 'Ta', 74: 'W', 75: 'Re', 76: 'Os', 77: 'Ir',
    78: 'Pt', 79: 'Au', 80: 'Hg', 81: 'Tl', 82: 'Pb',
    83: 'Bi', 84: 'Po', 85: 'At', 86: 'Rn', 87: 'Fr',
    88: 'Ra', 89: 'Ac', 90: 'Th', 91: 'Pa', 92: 'U',
    93: 'Np', 94: 'Pu', 95: 'Am', 96: 'Cm', 97: 'Bk',
    98: 'Cf', 99: 'Es', 100: 'Fm', 101: 'Md', 102: 'No',
    103: 'Lr', 104: 'Rf', 105: 'Db', 106: 'Sg', 107: 'Bh',
    108: 'Hs', 109: 'Mt', 110: 'Ds', 111: 'Rg', 112: 'Cn',
    113: 'Uut', 114: 'Fl', 115: 'Uup', 116: 'Lv',
    117: 'Uus', 118: 'Uuo'
}


class MAEExtractor:
    """
    Extracts the lowest energy conformer from a .maegz file.

    Macromodel conformer searches produce -out.maegz files containing
    all of the conformers found during the search and their energies
    and other data.

    Initializing this class with a :class:`.ConstructedMolecule` finds
    the ``-out.maegz`` file of that :class:`.ConstructedMolecule` and
    converts it to a ``.mae`` file. It then creates and additional
    ``.mae`` file holding only the lowest energy conformer found.

    Attributes
    ----------
    maegz_path : :class:`str`
        The path to the ``-out.maegz`` file generated by the macromodel
        conformer search.

    mae_path : :class:`str`
        The path to the ``.mae`` file holding the conformers generated
        by the macromodel conformer search.

    content : :class:`str`
        The content of the ``.mae`` file hodling all the conformers
        from the macromodel conformer search. This holds other data
        such as their energies too.

    energies : :class:`list`
        The :class:`list` has the form

        .. code-block:: python

            energies = [(0, 231.0), (1, 144.4), ...]

        Each :class:`tuple` holds the id and energy of every conformer
        in the ``.mae`` file, respectively.

    min_energy : :class:`float`
        The minimum energy found in the ``.mae`` file.

    path : :class:`str`
        The full path of the ``.mae`` file holding the extracted lowest
        energy conformer.

    """

    def __init__(self, run_name, n=1):
        self.maegz_path = f'{run_name}-out.maegz'
        self.maegz_to_mae()
        self.extract_conformers(n)

    def extract_conformers(self, n):
        """
        Creates ``.mae`` files holding the lowest energy conformers.

        Parameters
        ----------
        n : :class:`int`
            The number of conformers to extract.

        Returns
        -------
        None : :class:`NoneType`

        """

        for i in range(n):
            # Get the id of the lowest energy conformer.
            num = self.lowest_energy_conformers(n)[i][1]
            # Get the structure block corresponding to the lowest
            # energy conformer.
            content = self.content.split("f_m_ct")
            new_mae = "f_m_ct".join([content[0], content[num]])

            # Write the structure block in its own .mae file, named
            # after conformer extracted.
            if n == 1:
                # Write the structure block in its own .mae file, named
                # after conformer extracted.
                new_name = self.mae_path.replace(
                    '.mae',
                    f'EXTRACTED_{num}.mae'
                )
            else:
                new_name = self.mae_path.replace(
                    '.mae',
                    f'EXTRACTED_{num}_conf_{i}.mae'
                )

            with open(new_name, 'w') as mae_file:
                mae_file.write(new_mae)

            if i == 0:
                # Save the path of the newly created file.
                self.path = new_name

    def extract_energy(self, block):
        """
        Extracts the energy value from a ``.mae`` energy data block.

        Parameters
        ----------
        block : :class:`str`
            An ``.mae`` energy data block.

        Returns
        -------
        :class:`float`
            The energy value extracted from `block` or ``None`` if
            one is not found.

        """

        block = block.split(":::")
        for name, value in zip(block[0].split('\n'),
                               block[1].split('\n')):
            if 'r_mmod_Potential_Energy' in name:
                return float(value)

    def lowest_energy_conformers(self, n):
        """
        Returns the id and energy of the lowest energy conformers.

        Parameters
        ----------
        n : :class:`int`
            The number of lowest energy conformers to return.

        Returns
        -------
        :class:`list`
            A :class:`list` of the form

            .. code-block:: python

                returned = [(23, 123.3), (1, 143.89), (12, 150.6), ...]

            Where each :class:`tuple` holds the id and energy of the
            `n` lowest energy conformers, respectively.

        """

        # Open the .mae file holding all the conformers and load its
        # content.
        with open(self.mae_path, 'r') as mae_file:
            self.content = mae_file.read()
            # Split the content across curly braces. This divides the
            # various sections of the .mae file.
            content_split = re.split(r"[{}]", self.content)

        # Go through all the datablocks in the the .mae file. For each
        # energy block extract the energy and store it in the
        # `energies` list. Store the `index`  (conformer id) along with
        # each extracted energy.
        self.energies = []
        prev_block = deque([""], maxlen=1)
        index = 1
        for block in content_split:
            if ("f_m_ct" in prev_block[0] and
               "r_mmod_Potential_Energy" in block):
                energy = self.extract_energy(block)
                self.energies.append((energy, index))
                index += 1

            prev_block.append(block)

        # Selecting the lowest energy n conformers
        confs = sorted(self.energies)[:n]
        # Define the energy of the lowest energy conformer
        self.min_energy = confs[0][0]
        # Return a list with id and energy of the lowest energy
        # conformers.
        return confs

    def maegz_to_mae(self):
        """
        Converts the .maegz file to a .mae file.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.mae_path = self.maegz_path.replace('.maegz', '.mae')
        with gzip.open(self.maegz_path, 'r') as maegz_file:
            with open(self.mae_path, 'wb') as mae_file:
                mae_file.write(maegz_file.read())


def kill_macromodel():
    """
    Kills any applications left open as a result running MacroModel.

    Applications that are typically left open are
    ``jserver-watcher.exe`` and ``jservergo.exe``.

    Returns
    -------
    None : :class:`NoneType`

    """

    try:
        if os.name == 'nt':
            # In Windows, use the ``Taskkill`` command to
            # force a close on the applications.
            sp.run(
                ["Taskkill", "/IM", "jserver-watcher.exe", "/F"],
                stdout=sp.PIPE,
                stderr=sp.PIPE,
            )
            sp.run(
                ["Taskkill", "/IM", "jservergo.exe", "/F"],
                stdout=sp.PIPE,
                stderr=sp.PIPE,
            )

        if os.name == 'posix':
            sp.run(
                ["pkill", "jservergo"],
                stdout=sp.PIPE,
                stderr=sp.PIPE,
            )
            sp.run(
                ["pkill", "jserver-watcher"],
                stdout=sp.PIPE,
                stderr=sp.PIPE,
            )
    except Exception:
        pass


def mol_from_mae_file(mae_path):
    """
    Creates a ``rdkit`` molecule from a ``.mae`` file.

    Parameters
    ----------
    mol2_file : :class:`str`
        The full path of the ``.mae`` file from which an rdkit molecule
        should be instantiated.

    Returns
    -------
    :class:`rdkit.Mol`
        An ``rdkit`` instance of the molecule held in `mae_file`.

    """

    mol = rdkit.EditableMol(rdkit.Mol())
    conf = rdkit.Conformer()

    with open(mae_path, 'r') as mae:
        content = re.split(r'[{}]', mae.read())

    prev_block = deque([''], maxlen=1)
    for block in content:
        if 'm_atom[' in prev_block[0]:
            atom_block = block
        if 'm_bond[' in prev_block[0]:
            bond_block = block
        prev_block.append(block)

    labels, data_block, *_ = atom_block.split(':::')
    labels = [label for label in labels.split('\n')
              if not label.isspace() and label != '']

    data_block = [a.split() for a in data_block.split('\n') if
                  not a.isspace() and a != '']

    for line in data_block:
        line = [word for word in line if word != '"']
        if len(labels) != len(line):
            raise RuntimeError(('Number of labels does'
                                ' not match number of columns'
                                ' in .mae file.'))

        for label, data in zip(labels, line):
            if 'x_coord' in label:
                x = float(data)
            if 'y_coord' in label:
                y = float(data)
            if 'z_coord' in label:
                z = float(data)
            if 'atomic_number' in label:
                atom_num = int(data)

        atom_sym = periodic_table[atom_num]
        atom_coord = Point3D(x, y, z)
        atom_id = mol.AddAtom(rdkit.Atom(atom_sym))
        conf.SetAtomPosition(atom_id, atom_coord)

    labels, data_block, *_ = bond_block.split(':::')
    labels = [label for label in labels.split('\n')
              if not label.isspace() and label != '']
    data_block = [a.split() for a in data_block.split('\n')
                  if not a.isspace() and a != '']

    for line in data_block:
        if len(labels) != len(line):
            raise RuntimeError(('Number of labels does'
                                ' not match number of '
                                'columns in .mae file.'))

        for label, data in zip(labels, line):
            if 'from' in label:
                atom1 = int(data) - 1
            if 'to' in label:
                atom2 = int(data) - 1
            if 'order' in label:
                bond_order = str(int(data))
        mol.AddBond(atom1, atom2, bond_dict[bond_order])

    mol = mol.GetMol()
    mol.AddConformer(conf)
    return mol


def move_generated_macromodel_files(basename, output_dir):
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    for filename in iglob(f'{basename}*'):
        # Do not move the output_dir.
        if filename == output_dir:
            continue
        shutil.move(filename, f'{output_dir}/{filename}')


class XTBInvalidSolventError(Exception):
    ...


def is_valid_xtb_solvent(gfn_version, solvent_model, solvent):
    """
    Check if solvent is valid for the given GFN version.

    Parameters
    ----------
    gfn_version : :class:`int`
        GFN parameterization version. Can be: ``0``, ``1`` or ``2``.

    solvent_model : class:`str`
        Solvent model being used [1]_.

    solvent : :class:`str`
        Solvent being tested [1]_.

    Returns
    -------
    :class:`bool`
        ``True`` if solvent is valid.

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/gbsa.html

    """

    if gfn_version == 0:
        return False
    elif gfn_version == 1 and solvent_model == 'gbsa':
        valid_solvents = {
            'acetone', 'acetonitrile', 'benzene',
            'CH2Cl2'.lower(), 'CHCl3'.lower(), 'CS2'.lower(),
            'DMSO'.lower(), 'ether', 'H2O'.lower(),
            'methanol', 'THF'.lower(), 'toluene', 'water',
        }
    elif gfn_version == 1 and solvent_model == 'alpb':
        valid_solvents = {
            'acetone', 'acetonitrile', 'aniline', 'benzaldehyde',
            'benzene', 'CH2Cl2'.lower(), 'CHCl3'.lower(),
            'CS2'.lower(), 'dioxane', 'DMF'.lower(), 'DMSO'.lower(),
            'ether', 'ethylacetate', 'furane',
            'hexandecane', 'hexane', 'H2O'.lower(), 'nitromethane',
            'octanol', 'octanol (wet)', 'phenol', 'THF'.lower(),
            'toluene', 'water',
        }
    elif gfn_version == 2 and solvent_model == 'gbsa':
        valid_solvents = {
            'acetone', 'acetonitrile',
            'benzene', 'CH2Cl2'.lower(), 'CHCl3'.lower(),
            'CS2'.lower(), 'DMSO'.lower(),
            'ether', 'hexane', 'methanol', 'H2O'.lower(),
            'THF'.lower(), 'toluene', 'water',
        }
    elif gfn_version == 2 and solvent_model == 'alpb':
        valid_solvents = {
            'acetone', 'acetonitrile', 'aniline', 'benzaldehyde',
            'benzene', 'CH2Cl2'.lower(), 'CHCl3'.lower(),
            'CS2'.lower(), 'dioxane', 'DMF'.lower(), 'DMSO'.lower(),
            'ether', 'ethylacetate', 'furane',
            'hexandecane', 'hexane', 'H2O'.lower(), 'nitromethane',
            'octanol', 'octanol (wet)', 'phenol', 'THF'.lower(),
            'toluene', 'water',
        }
    return solvent in valid_solvents


def get_plane_normal(points):
    centroid = points.sum(axis=0) / len(points)
    return np.linalg.svd(points - centroid)[-1][2, :]


def has_h_atom(bond):
    """
    Check if a bond has a H atom.

    Parameters
    ----------
    bond : :class:`stk.Bond`
        Bond to test if it has a H atom.

    Returns
    -------
    :class:`bool`
        Returns `True` if bond has H atom.

    """

    if bond.get_atom1().get_atomic_number() == 1:
        return True
    if bond.get_atom2().get_atomic_number() == 1:
        return True

    return False


def has_metal_atom(bond, metal_atoms):
    """
    Check if a bond has a metal atom.

    Parameters
    ----------
    bond : :class:`stk.Bond`
        Bond to test if it has a metal atom.

    metal_atoms : :class:`.list` of :class:`stk.Atom`
        List of metal atoms.

    Returns
    -------
    :class:`bool`
        Returns `True` if bond has metal atom.

    """

    if bond.get_atom1() in metal_atoms:
        return True
    if bond.get_atom2() in metal_atoms:
        return True

    return False


def metal_atomic_numbers():

    return chain(range(21, 31), range(39, 49), range(72, 81))


def get_metal_atoms(mol):
    """
    Return a list of metal atoms in molecule.

    """

    metal_atoms = []
    for atom in mol.get_atoms():
        if atom.get_atomic_number() in metal_atomic_numbers():
            metal_atoms.append(atom)

    return metal_atoms


def get_metal_bonds(mol, metal_atoms):
    """
    Return a list of bonds in molecule that contain metal atoms.

    """

    metal_bonds = []
    ids_to_metals = []
    for bond in mol.get_bonds():
        if bond.get_atom1() in metal_atoms:
            metal_bonds.append(bond)
            ids_to_metals.append(bond.get_atom2().get_id())
        elif bond.get_atom2() in metal_atoms:
            metal_bonds.append(bond)
            ids_to_metals.append(bond.get_atom1().get_id())

    return metal_bonds, ids_to_metals


def to_rdkit_mol_without_metals(mol, metal_atoms, metal_bonds):
    """
    Create :class:`rdkit.Mol` with metals replaced by H atoms.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule to be optimized.

    metal_atoms : :class:`.list` of :class:`stk.Atom`
        List of metal atoms.

    metal_bonds : :class:`.list` of :class:`stk.Bond`
        List of bonds including metal atoms.

    Returns
    -------
    edit_mol : :class:`rdkit.Mol`
        RDKit molecule with metal atoms replaced with H atoms.

    """
    edit_mol = rdkit.EditableMol(rdkit.Mol())
    for atom in mol.get_atoms():
        if atom in metal_atoms:
            # In place of metals, add H's that will be constrained.
            # This allows the atom ids to not be changed.
            rdkit_atom = rdkit.Atom(1)
            rdkit_atom.SetFormalCharge(0)
        else:
            rdkit_atom = rdkit.Atom(atom.get_atomic_number())
            rdkit_atom.SetFormalCharge(atom.get_charge())
        edit_mol.AddAtom(rdkit_atom)

    for bond in mol.get_bonds():
        if bond in metal_bonds:
            # Do not add bonds to metal atoms (replaced with H's).
            continue
        edit_mol.AddBond(
            beginAtomIdx=bond.get_atom1().get_id(),
            endAtomIdx=bond.get_atom2().get_id(),
            order=rdkit.BondType(bond.get_order())
        )

    edit_mol = edit_mol.GetMol()
    rdkit_conf = rdkit.Conformer(mol.get_num_atoms())
    for atom_id, atom_coord in enumerate(mol.get_position_matrix()):
        rdkit_conf.SetAtomPosition(atom_id, atom_coord)
        edit_mol.GetAtomWithIdx(atom_id).SetNoImplicit(True)
    edit_mol.AddConformer(rdkit_conf)

    return edit_mol


def get_atom_distance(position_matrix, atom1_id, atom2_id):
    """
    Return the distance between two atoms.

    """

    distance = euclidean(
        u=position_matrix[atom1_id],
        v=position_matrix[atom2_id]
    )

    return float(distance)


def get_long_bond_ids(mol, reorder=False):
    """
    Return tuple of long bond ids in a ConstructedMolecule.

    """

    long_bond_ids = []
    for bond_infos in mol.get_bond_infos():
        if bond_infos.get_building_block() is None:
            if reorder:
                ba1 = bond_infos.get_bond().get_atom1().get_id()
                ba2 = bond_infos.get_bond().get_atom2().get_id()
                if ba1 < ba2:
                    ids = (
                        bond_infos.get_bond().get_atom1().get_id(),
                        bond_infos.get_bond().get_atom2().get_id(),
                    )
                else:
                    ids = (
                        bond_infos.get_bond().get_atom2().get_id(),
                        bond_infos.get_bond().get_atom1().get_id(),
                    )
            else:
                ids = (
                    bond_infos.get_bond().get_atom1().get_id(),
                    bond_infos.get_bond().get_atom2().get_id(),
                )
            long_bond_ids.append(ids)

    return tuple(long_bond_ids)


def calculate_dihedral(pt1, pt2, pt3, pt4):
    """
    Calculate the dihedral between four points in degrees.

    Uses Praxeolitic formula --> 1 sqrt, 1 cross product
    Output in range (-180 to 180).

    From: https://stackoverflow.com/questions/20305272/
    dihedral-torsion-angle-from-four-points-in-cartesian-
    coordinates-in-python
    (new_dihedral(p))

    """

    p0 = np.asarray(pt1)
    p1 = np.asarray(pt2)
    p2 = np.asarray(pt3)
    p3 = np.asarray(pt4)

    b0 = -1.0 * (p1 - p0)
    b1 = p2 - p1
    b2 = p3 - p2

    # normalize b1 so that it does not influence magnitude of vector
    # rejections that come next
    b1 /= np.linalg.norm(b1)

    # vector rejections
    # v = projection of b0 onto plane perpendicular to b1
    #   = b0 minus component that aligns with b1
    # w = projection of b2 onto plane perpendicular to b1
    #   = b2 minus component that aligns with b1
    v = b0 - np.dot(b0, b1) * b1
    w = b2 - np.dot(b2, b1) * b1

    # angle between v and w in a plane is the torsion angle
    # v and w may not be normalized but that's fine since tan is y/x
    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)
    return np.degrees(np.arctan2(y, x))


def vector_angle(vector1, vector2):
    """
    Returns the angle between two vectors in radians.
    Parameters
    ----------
    vector1 : :class:`numpy.ndarray`
        The first vector.
    vector2 : :class:`numpy.ndarray`
        The second vector.
    Returns
    -------
    :class:`float`
        The angle between `vector1` and `vector2` in radians.
    """

    if np.all(np.equal(vector1, vector2)):
        return 0.

    numerator = np.dot(vector1, vector2)
    denominator = np.linalg.norm(vector1) * np.linalg.norm(vector2)
    # This if statement prevents returns of NaN due to floating point
    # inaccuracy.
    term = numerator/denominator
    if term >= 1.:
        return 0.0
    if term <= -1.:
        return np.pi
    return np.arccos(term)


def calculate_angle(pt1, pt2, pt3):
    """
    Calculate the angle between three points in degrees.

    """

    v1 = pt1 - pt2
    v2 = pt3 - pt2
    return np.degrees(vector_angle(v1, v2))


def get_torsion_info_angles(mol, torsion_info):
    """
    Get the angles for torsion_info in mol.

    The first angle returned is torsion angle in the
    :class:`stk.ConstructedMolecule`.
    The second angle returned is torsion angle in the
    :class:`stk.BuildingBlock`.
    Both angles are in degrees.

    A :class:`stko.MatchedTorsionCalculator` should yield torsions
    such that the two angles returned are the same.

    Parameters
    ----------
    mol : :class:`.ConstructedMolecule`
        The :class:`.ConstructedMolecule` for which angles are
        computed.

    torsion_info : TorsionInfo
        Specifies the torsion for which angles will be computed.

    Returns
    -------
    angle : :class:`float`, bb_angle : :class:`float`

    """

    torsion = torsion_info.get_torsion()
    angle = calculate_dihedral(
        pt1=tuple(
            mol.get_atomic_positions(
                torsion.get_atom_ids()[0]
            )
        )[0],
        pt2=tuple(
            mol.get_atomic_positions(
                torsion.get_atom_ids()[1]
            )
        )[0],
        pt3=tuple(
            mol.get_atomic_positions(
                torsion.get_atom_ids()[2]
            )
        )[0],
        pt4=tuple(
            mol.get_atomic_positions(
                torsion.get_atom_ids()[3]
            )
        )[0],
    )
    bb_torsion = torsion_info.get_building_block_torsion()
    if bb_torsion is None:
        bb_angle = None
    else:
        bb_angle = calculate_dihedral(
            pt1=tuple(
                torsion_info.get_building_block().get_atomic_positions(
                    bb_torsion.get_atom_ids()[0]
                )
            )[0],
            pt2=tuple(
                torsion_info.get_building_block().get_atomic_positions(
                    bb_torsion.get_atom_ids()[1]
                )
            )[0],
            pt3=tuple(
                torsion_info.get_building_block().get_atomic_positions(
                    bb_torsion.get_atom_ids()[2]
                )
            )[0],
            pt4=tuple(
                torsion_info.get_building_block().get_atomic_positions(
                    bb_torsion.get_atom_ids()[3]
                )
            )[0],
        )
    return angle, bb_angle


def get_atom_maps(mol):
    """
    Get atom maps from building blocks to constructude molecule.

    Returns a dictionary of dictionaries from atom id (in building
    block) to constructed molecule atom, indexed by building block id.

    Parameters
    ----------
    mol : :class:`.ConstructedMolecule`
        The :class:`.ConstructedMolecule` for which atom maps are
        desired.

    """
    atom_maps = defaultdict(dict)
    for atom_info in mol.get_atom_infos():
        bb_atom_id = atom_info.get_building_block_atom().get_id()
        atom_maps[atom_info.get_building_block_id()][
            bb_atom_id
        ] = atom_info.get_atom()
    return atom_maps



import stko
import numpy as np


def test_universe(molecule):
    """
    Test :meth:`.MDAnalysis.get_universe`.

    Parameters
    ----------
    molecule : :class:`stk.Molecule`
        The molecule to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    result = stko.MDAnalysis().get_universe(molecule)

    assert result.atoms.n_atoms == molecule.get_num_atoms()

    for atom, stk_atom in zip(result.atoms, molecule.get_atoms()):
        assert atom.ix == stk_atom.get_id()

    for pos, stk_pos in zip(
        result.atoms.positions, molecule.get_position_matrix()
    ):
        assert np.allclose(pos, stk_pos)



"""
Orca Results
============

#. :class:`.OrcaResults`

Results class for the output of Orca.

"""

from .results import Results
from ..extractors import OrcaExtractor


class OrcaResults(Results):
    """
    Results class containing molecule Orca properties.

    """

    def __init__(
        self,
        generator,
        output_file,
        extractor=OrcaExtractor,
    ):

        # Run calculation.
        next(generator)
        self._extractor = extractor(output_file=output_file)

    def get_total_energy(self):
        return (self._extractor.total_energy, 'a.u.')



from ..utilities import is_equivalent_atom


def test_with_id(positioned_atom, id):
    """
    Test :meth:`.PositionedAtom.with_id`.

    Parameters
    ----------
    atom : :class:`.PositionedAtom`
        The atom to test.

    id : :class:`int`
        The correct id of the new atom.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a clone to ensure that "atom" is not changed by the test.
    before = positioned_atom.clone()
    _test_with_id(positioned_atom, id)
    is_equivalent_atom(before, positioned_atom)


def _test_with_id(positioned_atom, id):
    """
    Test :meth:`.PositionedAtom.with_id`.

    Parameters
    ----------
    atom : :class:`.PositionedAtom`
        The atom to test.

    id : :class:`int`
        The correct id of the new atom.

    Returns
    -------
    None : :class:`NoneType`

    """

    new_atom = positioned_atom.with_id(id)
    assert new_atom is not positioned_atom
    assert new_atom.__class__ is positioned_atom.__class__
    assert new_atom.get_id() == id
    assert new_atom.get_charge() == positioned_atom.get_charge()
    atm_num = positioned_atom.get_atomic_number()
    assert new_atom.get_atomic_number() == atm_num
    assert new_atom.get_position() == positioned_atom.get_position()
